from typing import Optional, List, Dict
from logic import queries
import uuid
from pydantic import BaseModel


class Data(BaseModel):
    x: List[float]
    min: List[float]
    max: List[float]
    mean: List[float]
    count: List[int]


def insert_event(
    user_id: uuid.UUID,
    timestamp: float,
    name: str,
    period_min: float,
    period_max: float,
    period_mean: float,
    period_count: float,
) -> None:
    """
    Insert profiling event.

    Args:
        user_id: Which user
        timestamp: UNIX timestamp
        name: Event name
        period_min: Min seconds elapsed,
        period_max: Max seconds elapsed,
        period_mean: Mean seconds elapsed,
        period_count: Number of data points,
    """
    queries.insert_event(
        user_id,
        timestamp,
        name,
        period_min,
        period_max,
        period_mean,
        period_count,
    )


def retrieve_data(user_id: uuid.UUID, start: float, end: Optional[float]) -> Dict[str, Data]:
    """
    Retrieve data for a given user between a given period.

    Args:
        user_id: Which user
        start: Start time
        end: End time

    Return:
        Dictionary where key is task_name, value is data container.
    """
    data: Dict[str, Data] = {}
    for timestamp, name, period_min, period_max, period_mean, period_count in queries.retrieve_data(user_id, start, end):
        if name not in data:
            data[name] = Data(x=[], min=[], max=[], mean=[], count=[])

        data[name].x.append(timestamp)
        data[name].min.append(period_min)
        data[name].max.append(period_max)
        data[name].mean.append(period_mean)
        data[name].count.append(period_count)

    return data



import os
import time
from typing import Any, List, Optional, Tuple

import config
import sqlite3
from commons.logging import logger


def _upgrade_schema(version: int) -> None:
    """
    Upgrade schema by one version.

    Args:
        version: Next version

    Raises:
        FileNotFoundError: If schema with version number is not found
    """
    target_schema = os.path.join(os.path.dirname(__file__), f"{version}.sql")

    if not os.path.exists(target_schema):
        raise FileNotFoundError(f"\tSchema does not exist: {target_schema}")

    with open(target_schema, "r") as patch:
        logger.info(f"\tUpgrading schema to version: {version}")

        execute_script(patch.read())

        execute(
            """
                INSERT INTO schema_version (version, date)
                VALUES (?, ?);
            """,
            (version, int(time.time())),
        )


path = os.path.join(config.PROJECT_ROOT, "data")
if not os.path.exists(path):
    os.makedirs(path)

def _open_connection() -> sqlite3.Connection:
    return sqlite3.connect(os.path.join(path, "jetfuel.db"))


def execute_script(statement: str) -> None:
    conn = _open_connection()

    cursor = conn.cursor()
    cursor.executescript(statement)
    cursor.close()

    conn.commit()
    conn.close()


def execute(statement: str, args: Optional[Tuple[Any, ...]] = None) -> None:
    conn = _open_connection()

    cursor = conn.cursor()
    cursor.execute(statement, args)
    cursor.close()

    conn.commit()
    conn.close()


def retrieve_one(statement: str, args: Optional[Tuple[Any, ...]] = None) -> Tuple[Any]:
    conn = _open_connection()

    cursor = conn.cursor()
    result = cursor.execute(statement, args).fetchone()
    cursor.close()

    conn.commit()
    conn.close()

    return result


def retrieve_all(statement: str, args: Optional[Tuple[Any, ...]] = None) -> List[Tuple[Any]]:
    conn = _open_connection()

    cursor = conn.cursor()
    result = cursor.execute(statement, args).fetchall()
    cursor.close()

    conn.commit()
    conn.close()

    return result


def init() -> None:
    """
    Upgrades SQL Schema using incremental scripts. New setups will start
    with schema version 0, and incrementally build required schema.

    Raises:
        RuntimeError: If detected schema is higher than required.
    """
    logger.info("Initializing SQL")

    version = -1
    try:
        (version,) = retrieve_one("SELECT MAX(version) FROM schema_version;")
    except:  # No schema detected
        pass

    while version < config.SQL_VERSION:
        version += 1
        _upgrade_schema(version)



from fastapi import Depends, FastAPI, HTTPException, status
from fastapi.security import HTTPBasicCredentials, HTTPBearer
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi.responses import RedirectResponse, HTMLResponse
from pydantic import BaseModel
from commons.logging import exception_handler, logger
from logic import auth, data
from typing import List, Dict, Optional
import config


app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
    allow_credentials=True,
)


# Bearer token for client events
security = HTTPBearer()


class EventBody(BaseModel):
    # Represents a given period of data collection
    class TaskPeriod(BaseModel):
        class Stats(BaseModel):
            min: float
            max: float
            mean: float
            count: int

        name: str
        stats: Stats

    timestamp: float
    periods: List[TaskPeriod]


class DataResponse(BaseModel):
    data: Dict[str, data.Data]


@app.post("/v1/commit")
@exception_handler()
def post_commit(event: EventBody, credentials: HTTPBasicCredentials = Depends(security)) -> None:
    """
    Receive profiler event from client.

    Args:
        event: Event body containing name and seconds elapsed.
        credentials: Bearer with API Token
    """
    api_token = credentials.credentials
    user_id = auth.verify_api_token(api_token)
    if not user_id:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Unauthorized",
        )

    for period in event.periods:
        data.insert_event(
            user_id,
            event.timestamp,
            period.name,
            period.stats.min,
            period.stats.max,
            period.stats.mean,
            period.stats.count,
        )

    return "OK"


@app.get("/v1/data")
@exception_handler()
def get_data(
    start: float,
    end: Optional[float] = None,
    credentials: HTTPBasicCredentials = Depends(security),
) -> DataResponse:
    """
    Retrieve data for web client to display as graphs.

    If user contained in token doesn't exist, we will automatically create a new one.

    Args:
        start: Start UNIX time
        end: Optional end UNIX time
        credentials: Bearer with Auth0 Token or "default"
    """
    user_token = credentials.credentials
    user_id = auth.verify_user_token(user_token)
    if not user_id:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Unauthorized",
        )

    data_ = data.retrieve_data(user_id, start, end)

    return DataResponse(data=data_)


# Serve Frontend
@app.get("/")
async def redirect_webapp():
    return RedirectResponse("/index.html")


app.mount("/", StaticFiles(directory="frontend"), name="frontend")



@app.on_event("startup")
def startup() -> None:
    logger.info("")
    logger.info("Jetfuel Server")
    logger.info("--------------")
    logger.info("")

    # Add Default User
    if config.DEFAULT_USER:
        try:
            user_id = auth.create_user(email="default")
            auth.generate_api_token(user_id=user_id, api_token="default")
        except:
            # Default user already exists
            pass

    logger.info("")
    logger.info("READY")
    logger.info("")



"""
 Calculate stats (min, max, mean, count), then fire event to Jetfuel-Server
                                    ^
                                    |
           Event Cache (stores raw performance profiling events)
                                    ^
                                    |
                              Profiler Class
                                    ^
                                    |
        Syntactic Sugar (context manager, function decorator, etc.)
"""

from .state import init as init
from .profiler import Profiler as Profiler
from .profiler import profiler as profiler
from .profiler import start as start
from .demo import demo as demo



from typing import Dict, Union, List
from threading import Thread, Lock
from .cache import Cache
import requests
import time


_url: str = ""
_token: str = ""
_resolution: float = 5.0  # seconds
_caches: Dict[str, Cache] = {}
_commit_lock = Lock()


def init(url: str, token: str = "default", resolution: float = 5.0) -> None:
    """
    Initialize Jetfuel.

    Args:
        url: URL of jetfuel-server
        token: Login token
        resolution: Data collection resolution
    """
    global _url, _token, _resolution

    _url = url
    _token = token
    _resolution = resolution

    # Start data sending thread
    t = Thread(target=_commit)
    t.start()


def update(profiler_name: str, elapsed_time: float) -> None:
    """
    Receive update for a given profiler.

    Args:
        profiler_name: Which profiler
        elapsed_time: Time elapsed for task
    """
    global _caches, _url, _token, _commit_lock

    if profiler_name not in _caches:
        with _commit_lock:
            _caches[profiler_name] = Cache(profiler_name)

    _caches[profiler_name].update(elapsed_time)


def _commit() -> None:
    """
    Commit data to backend.
    """
    global _resolution

    while True:
        time.sleep(_resolution)

        with _commit_lock:
            periods: List[Dict[str, Union[str, Dict[str, Union[int, float]]]]] = []
            profiler_names = list(_caches.keys())

            try:
                # Acquire locks and calculate stats body
                for profiler_name in profiler_names:
                    cache = _caches[profiler_name]

                    # Ensure that cache updates, stat_calculation, and flushes are thread-safe
                    cache.lock.acquire()

                    # Calculate stats and add to request body
                    stats = cache.stats()
                    if stats is not None:
                        period_min, period_max, period_mean, period_count = stats
                        periods.append({
                            "name": profiler_name,
                            "stats": {
                                "min": period_min,
                                "max": period_max,
                                "mean": period_mean,
                                "count": period_count,
                            }
                        })

                # Commit to backend
                if len(periods) > 0:
                    requests.post(
                        f"{_url}/v1/commit",
                        json={
                            "timestamp": time.time(),
                            "periods": periods,
                        },
                        headers={
                            "Authorization": f"Bearer {_token}"
                        }
                    )

                # Flush cache data
                for cache in _caches.values():
                    cache.flush()
            except:
                pass
            finally:
                # Release locks
                for profiler_name in profiler_names:
                    cache = _caches[profiler_name]
                    cache.lock.release()



from typing import List, Tuple, Optional
from threading import Lock


class Cache:
    def __init__(self, name: str) -> None:
        """
        Initialize thread-safe cache for a given profiler type.

        Args:
            name: Profiler name
            url: URL of Jetfuel-Server
            token: Auth token to use
        """
        self.name = name
        self.data: List[float] = []
        self.last_commit: float = 0
        self.lock = Lock()
    
    def stats(self) -> Optional[Tuple[float, float, float, int]]:
        """
        Calculate stats from data.

        Returns:
            period_min
            period_max
            period_mean
            period_count
        """
        if len(self.data) == 0:
            return

        # Calculate stats
        period_min = min(self.data)
        period_max = max(self.data)
        period_count = len(self.data)
        period_mean = sum(self.data) / period_count

        return period_min, period_max, period_mean, period_count

    def flush(self) -> None:
        """
        Flush data in cache.
        """
        # Flush data
        self.data = []


    def update(self, elapsed_time: float) -> None:
        """
        Update data.

        Args:
            elapsed_time: Time elapsed for a given Profiler.
        """
        with self.lock:
            self.data.append(elapsed_time)



from typing import Callable, Any
from functools import wraps
from .state import update
import time


class Profiler:
    """
    Base Profiler that is used by all other convenience functions.
    """

    def __init__(self, name: str) -> None:
        """
        Initialize Profiler.

        Args:
            name: Name of task being profiled.
        """
        self.name = name
    

    def start(self) -> None:
        """
        Start profiler.
        """
        self.start_time = time.perf_counter()


    def stop(self) -> None:
        """
        Stop profiler.
        """
        elapsed_time = time.perf_counter() - self.start_time
        update(self.name, elapsed_time)


    def __enter__(self):
        """
        Start profiler as a context manager.
        """
        self.start()


    def __exit__(self, type, value, traceback):
        """
        Stop profiler context manager.
        """
        self.stop()


def start(name: str) -> Profiler:
    """
    Start a profiler.

    Args:
        name: Task name

    Returns:
        Profiler
    """
    profiler = Profiler(name=name)
    profiler.start()

    return profiler


def profiler(name: str) -> Callable[..., Callable[..., Any]]:
    """
    Profiler decorator.

    Args:
        name: Task name

    Returns:
        Function wrapper
    """
    def wrapper(f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            with Profiler(name=name):
                return f(*args, **kwargs)
        return wrapped
    return wrapper



import logging
import logging.handlers
import os
import sys
import traceback
from functools import wraps
from typing import Any, Callable

from coloredlogs import ColoredFormatter

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel("INFO")

if not os.path.exists("/var/log/jetfuel"):
    os.makedirs("/var/log/jetfuel")
log_file = os.path.join(os.path.abspath(os.sep), "/var/log/jetfuel/root.log")

level_colors = {
    "debug": {"color": "green"},
    "info": {},
    "warning": {"color": "yellow"},
    "error": {"color": "red"},
    "critical": {"bold": True, "color": "white", "background": "red"},
}

formatter: logging.Formatter = ColoredFormatter(
    fmt="%(levelname)s:     %(message)s", level_styles=level_colors
)

file_handler: logging.Handler = logging.FileHandler(log_file)
file_handler.setFormatter(formatter)
file_handler.setLevel(logging.DEBUG)

stream_handler: logging.Handler = logging.StreamHandler(sys.__stdout__)
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.DEBUG)

logger.addHandler(file_handler)
logger.addHandler(stream_handler)


def exception_handler() -> Callable[..., Callable[..., Any]]:
    """
    Python decorator used to catch FastAPI errors and print traceback.
    """

    def wrapper(f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            try:
                return f(*args, **kwargs)
            except BaseException as e:
                for traceback_line in traceback.format_exc().split("\n"):
                    logger.error(f"{traceback_line}")
                raise e

        return wrapped

    return wrapper



"""
All queries that are made to external databases are defined here.
This allows us to add new database backends in the future in desired.
"""
import config
from typing import Optional, Tuple, List
from commons.database import sqlite
import uuid


if config.SQL_MODE == "sqlite":
    sqlite.init()


def create_user(email: str, user_id: uuid.UUID) -> None:
    """
    Create a user.

    Args:
        email: Unique user email
        user_id: Unique User ID
    """
    if config.SQL_MODE == "sqlite":
        sqlite.execute(
            """
                INSERT INTO users (user_id, email)
                VALUES (?, ?);
            """,
            (str(user_id), email)
        )


def get_user_id(email: str) -> Optional[uuid.UUID]:
    """
    Retrieve a user.

    Args:
        email: Unique user email
    """
    if config.SQL_MODE == "sqlite":
        try:
            (user_id_str,) = sqlite.retrieve_one(
                """
                    SELECT user_id
                    FROM users
                    WHERE email=?;
                """,
                (email,)
            )

            return uuid.UUID(user_id_str)
        except:
            return None


def update_api_token(user_id: uuid.UUID, api_token: str) -> None:
    """
    Update a user's api_token.

    Args:
        user_id: Which user
        api_token: User's api_token
    """
    if config.SQL_MODE == "sqlite":
        sqlite.execute(
            """
                UPDATE users SET api_token=? WHERE user_id=?;
            """,
            (api_token, str(user_id))
        )


def verify_api_token(api_token: str) -> Optional[uuid.UUID]:
    """
    Verify an api_token. Returns None if api_token is invalid.

    Returns:
        user_id
    """
    try:
        if config.SQL_MODE == "sqlite":
            (user_id,) = sqlite.retrieve_one(
                """
                    SELECT user_id FROM users WHERE api_token=?;
                """,
                (api_token,)
            )

            return user_id
    except:
        return None
    


def insert_event(
    user_id: uuid.UUID,
    timestamp: float,
    name: str,
    period_min: float,
    period_max: float,
    period_mean: float,
    period_count: float,
) -> None:
    """
    Insert profiling event.

    Args:
        user_id: Which user
        timestamp: UNIX timestamp
        name: Event name
        period_min: Min seconds elapsed,
        period_max: Max seconds elapsed,
        period_mean: Mean seconds elapsed,
        period_count: Number of data points,
    """
    if config.SQL_MODE == "sqlite":
        sqlite.execute(
            """
                INSERT INTO events (user_id, timestamp, name, period_min, period_max, period_mean, period_count)
                VALUES (?, ?, ?, ?, ?, ?, ?);
            """,
            (user_id, timestamp, name, period_min, period_max, period_mean, period_count)
        )


def retrieve_data(
    user_id: uuid.UUID, start: float, end: Optional[float]
) -> List[Tuple[float, str, float, float, float, int]]:
    """
    Retrieve data for a given user between a given period.

    Args:
        user_id: Which user
        start: Start time
        end: End time

    Return:
        List:
            timestamp
            name
            period_min
            period_max
            period_mean
            period_count
    """
    if config.SQL_MODE == "sqlite":
        result: List[Tuple[float, str, float, float, float, int]] = []

        if end is None:
            result = sqlite.retrieve_all(
                """
                    SELECT timestamp, name, period_min, period_max, period_mean, period_count
                    FROM events
                    WHERE user_id=? AND timestamp>=?;
                """,
                (str(user_id), start)
            )
        else:
            result = sqlite.retrieve_all(
                """
                    SELECT timestamp, name, period_min, period_max, period_mean, period_count
                    FROM events
                    WHERE user_id=? AND timestamp>=? AND timestamp<=?;
                """,
                (str(user_id), start, end)
            )


        return result



import setuptools

setuptools.setup(
    name="jetfuel",
    version="0.0.12",
    author="Jared Zhao",
    description="Python Performance Profiling for Production",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jetfuel-dev/jetfuel",
    license="Apache 2.0",
    python_requires=">=3.6",
    install_requires=[
        "requests >= 2.0.0",
    ],
    packages=["jetfuel"],
)



from typing import Optional
from logic import queries
import config
import uuid


def create_user(email: str) -> uuid.UUID:
    """
    Create user.
    """
    user_id = uuid.uuid4()
    queries.create_user(email=email, user_id=user_id)

    return user_id


def generate_api_token(user_id: uuid.UUID, api_token: Optional[str] = None) -> None:
    """
    Generate a api_token for a user.

    Args:
        user_id: Which user
        api_token: Optional default api_token
    """
    if api_token is None:
        # TODO: Add way to generate new api_token
        raise NotImplementedError()

    queries.update_api_token(user_id, api_token)


def verify_api_token(api_token: str) -> Optional[uuid.UUID]:
    """
    Verify an api_token. Returns None if api_token is invalid.

    Returns:
        user_id
    """
    return queries.verify_api_token(api_token)


def verify_user_token(user_token: str) -> Optional[uuid.UUID]:
    """
    Verify user token (Auth0 or "default").

    Args:
        user_token: User's token

    Returns:
        user_id
    """
    email: Optional[str] = None

    if config.DEFAULT_USER and user_token == "default":
        email = user_token

    # TODO: Add Auth0 email verification

    assert email is not None

    return queries.get_user_id(email)



def demo() -> None:
    """
    Run demo. Jetfuel must be initialized first.
    """
    from .profiler import Profiler
    from concurrent.futures import ThreadPoolExecutor
    import random
    import time

    tasks = [
        "Workspace Creation (New Customer)",
        "Dataset Join - 259 GB",
        "ETL Snowflake -> Amazon Glacier",
        "CI/CD full build",
        "Terraform Prod Deployment",
    ]

    demo_running = True

    def run_task(task) -> None:
        while demo_running:
            with Profiler(task):
                time.sleep(random.random())

    try:
        with ThreadPoolExecutor(max_workers=len(tasks)) as executor:
            for task in tasks:
                executor.submit(run_task, task=task)
    except:
        # Catch keyboard interrupt (or any other exception), then stop demo
        demo_running = False
